package IceBearSoft::Linkchecker;

# Link checker package (discovers broken links)
# (C) Z. Wagner -- Ice Bear Soft, 31 Dec 1998

# Modifications:
# 31 Jan 1999 v. 1.0a: honours robots.txt and
#                      <META name="robots" content="... nofollow ...">
#                      Status other than 200 is displayed red in normal display
#                      Checks redirection count (must not exceed 5)
#                      Handles <META http-equiv="refresh" ...>
#                      Test of dead browser
#                      Warning if <IMG SRC=...> not of type "image/..."
#                      More links to the Help Window
#  9 Jun 1999 v. 1.0b: adapted to the new version of Zwebfun (accepts HASHREF)
#  1 Jan 2000 v. 1.0c: Bug fixed (some local pages were pushed repeatedly)
#                      Fixed bug in ZWsgml usage
#                      Checking is forbidden if we cannot connect when testing robots.txt
#                      Incremental checking
#                      Almost all functionality implemented here (not in linkcheck.pl)
#                      Enhanced comments in Linkchecker.pm
#                      Added support for <BODY BACKGROUND=...> and warning if it is not
#                      of type "image/..."
# 31 Jul 2002 v. 1.0d: IP address of the default proxy taken from $ENV{'DEFAULT_PROXY'}
#  5 Sep 2003          Fixed small bug in CheckExists, version number intentionally unchanged
# 13 Dec 2004 v. 1.0e: Forgotten </font> added
#  5 Mar 2005 v. 1.1:  The file placed into CVS

# Usage:
# 1. Create an object via: $checker = new IceBearSoft::Linkchecker;
# 2. Modify variables (see below "new")
# 3. Call the go method

use IceBearSoft::Zwebfun;
use IceBearSoft::ZWsgml;
use IceBearSoft::ZWurl(abs_url, split_url, merge_url, print_ref, http_url, cgi_parse, url_encode);
require Exporter;
@ISA = qw(Exporter);

# Create a new link checker object, no arguments
#
# Variables:
# ==========
# Name           Type            Description
# ---------------------------------------------------------------------------------
# http           reference       IceBearSoft::Zwebfun object
# User-Agent     string          Description of User Agent, do not change it!
# Loaded         hash ref        Hash with information about loaded resources
# Temporary-File string          Name of the temporary file, do not change it!
# Keep-Alive     boolean         Use persistent connection
# Version        numeric string  HTTP version (will be passed to http), default 1.1
# Proxy          URL or IP       Proxy server, will be passed to http
# Debug-Response boolean         Display full response headers
# This           string          hostname of the first object
# ID             number          Number of the URL, do not change it!
# Summary        hash ref        Errors and warning summary table, do not change!
# ---------------------------------------------------------------------------------
#
# You can change HTTP timeout by:
# $checker->{'http'}->{'read-timeout'} = 600;

# ---------------------------------------------------------------------------------
# Contents of Loaded:
# ===================
# This is a hash indexed by URLs of loaded objects. Each element is again a hash
# reference with following fields:
#
# Body -- hash filled by the Get function
#    Form -- contains URLs of the forms, it is a hash indexed by URLs and its contents
#            is the request method
#    Frame -- hash containing information about frames indexed by URLs, each element
#             contains the name of the target window
#    Href -- hash containing hypertext references indexed by URLs, each element
#            contains the type of reference (A, IMG, ...)
#    Meta -- hash of <META ...> tags, only refresh is interesting
#    Name -- list of labels (<a name=...>), all are preceded by # and their value is 0
# Checked -- YES means that the page was already checked, PUSHED means that it was
#            pushed to the array of local pages to be checked
# Error-Message -- copy of Error-Message from Response (see below)
# Method -- request method used for loading the object
# Previous -- YES means that the object was loaded by previous run
# Response -- hash of response headers. The following fields are added (by IceBearSoft::Zwebfun):
#    Error-Message -- reason why the object was not loaded (host not found etc.)
#    Previous -- YES means that the object was loaded by previous run
#    Protocol-Type -- information retrieved from the status line, usually contains HTTP
#    Protocol-Version -- protocol version from the status line
#    Reason-Phrase -- text message from the status line
#    Status -- the status code returned on the status line
#    Status-Line -- the full contents of the status line (the first line of the response)
#    Sub-Status -- minor version of status, some servers send e.g. 404.1
#    file -- file used to store the body (GET only). It is deleted after parsing
#    proxy -- proxy used to download the object
#    url -- URL of the object (required by IceBearSoft::Zwebfun)
# Status -- copy of Status from Response
# Type -- copy of Content-Type from Response
# URL -- URL of the object, must be the same as url in Response and the key of the whole
#        hash. It is used in cases where the pointer to the hash is known but the key is
#        not supplied.

sub new ($*) {
my $class = shift;
my $this = {};
$this->{'http'} = new IceBearSoft::Zwebfun;
$this->{'User-Agent'} = 'LinkChecker/1.2';
$this->{'Keep-Alive'} = 0;
$this->{'Debug-Response'} = 0;
$this->{'Loaded'} = {};
$this->{'User-Headers'} = {};
my $temp = $ENV{'TEMP'};
if ($temp) {
  $temp =~ s!\\!/!g;
  $temp .= '/' unless $temp =~ /\/$/;
}
else {
  $temp = 'C:/';
}
$this->{'Temporary-File'} = sprintf "$temp%d.%d", time(), $$;
$this->{'ID'} = 0;
$this->{'LinkChecker-Help-Window'} =
        'http://'.$ENV{'SERVER_NAME'}.'/webtools/linkcheck-help.html';
$this->{'Summary'} = {};
return bless $this, $class;
}

# Put an error into a summary table and show info
# ===============================================

sub SummarizeError {
my ($this, $errcode, $id) = @_;
#$this->dbgsum();
my $summary = $this->{'Summary'};
my $previous = '';
if (exists $summary->{$errcode}) {
  my $err = $summary->{$errcode};
  $previous = $err->{'last'};
  $err->{'last'} = $id;
  $err->{'count'} = 1 + $err->{'count'};
#  print "<P>'$errcode' exists, previous: $previous.\n" if $this->{'debug'};
} else {
#  print "<P>'$errcode' does not exist.\n" if $this->{'debug'};
  my %err = (
    'count' => 1, 'last' => $id
  );
  $summary->{$errcode} = \%err;
}
#$this->dbgsum();
if ($previous) {
  print "<P><font color=red>Previous occurence:</font> <a href=\"#$previous\">$previous</a>\n";
} else {
  print "<P><font color=red>First occurence,</font> go to the <a href=\"#summary\">summary table</a>\n";
}
}

sub dbgsum {
my $this = shift;
return unless $this->{'debug'} eq 'on';
my %hash = %$this->{'Summary'};
print "<P><HR><B>";
print_ref $this->{'Summary'};
print "</B>\n";
print_ref \%hash;
print "<HR>\n";
}

# Presents a link to the Help window
# ==================================

sub HelpWindow ($) {
my $this = shift;
my $helpwin = $this->{'LinkChecker-Help-Window'};
print "<P><a target=\"HELP\" href=\"$helpwin\">Help Window</a>\n"
      or die "Browser died";
}

# Displays "too many redirections"
# ================================

sub RedirError ($$$) {
my ($this, $id, $url) = @_;
print "<H3><a name=\"$id\">$id</a> <A TARGET=\"TEST\" HREF=\"$url\">$url</A></H3>\n";
print '<P><font color="#FF0000">Too many redirections!</font>';
print "<BR>The redirection chain is too long, it is probably an infinite loop\n"
      or die "Browser died";
$this->SummarizeError('Too many redirection', $id);
}

# Gets \%url and sets Error-Message if loading forbidden by robots.txt
# ====================================================================

sub CheckRobots ($$) {
my ($this, $url) = @_;
return if !$url->{'URL'};
my %url = split_url $url->{'URL'};
return if $url{'scheme'} ne 'http' or !$url{'host'};
my $object = $url{'object'};
return if $object eq '/robots.txt';
$robots = "http://$url{'host'}/robots.txt";
my $answer = $this->Head($robots);  # Get robots.txt
return if !$answer;
$answer = $this->Get($robots);  # Get robots.txt
return if !$answer;
my $status = $answer->{'Status'};
if (!$status || $status>=500 || exists($answer->{'Error-Message'}) ||
                !exists($answer->{'Response'})) {
  # This make verification faster if connection failed during download of robots.txt
  my $errmsg = $answer->{'Error-Message'};
  $errmsg = $answer->{'Response'}->{'Reason-Phrase'} if !$errmsg;
  $errmsg = 'Cannot connect to server' if !$errmsg;
  $url->{'Error-Message'} = $url->{'Response'}->{'Error-Message'} = $errmsg;
}
$disallow = $answer->{'Robots'};   # List of not allowed objects
if (!$disallow) {
  return unless $answer->{'Status'} == 200;
  my $response = $answer->{'Response'};
  return unless $response;  # robots.txt not found
  return unless $response->{'Content-Type'} =~ /^text\/plain/;
  my $file = $response->{'file'};
  return unless $file;
  my @disallow;
  open(ROBOTS, $file);   # Parse robots.txt
  my $ctl = 0;
  WHILE: while ($line = <ROBOTS>) {
    $line =~ /^\s*(.*)\s*$/;  $line = $1;
    SWITCH: {
      if (!$ctl) {
        next WHILE if $line =~ /^$/;
        my $agent = '';
        if ($line =~ /^user-agent:\s*/i) {
          $agent = $';
        }
        $ctl = ($agent eq '*' or lc($agent) eq lc('LinkChecker')) ? 1 : -1;
        last SWITCH;
      }
      last SWITCH if $ctl < 0;
      if ($ctl > 0) {
        if ($line =~ /^disallow:\s*/i) {
          push @disallow, $';
        }
        last SWITCH;
      }
    }
  }
  close ROBOTS;
  $disallow = $answer->{'Robots'} = \@disallow;
}
return unless $disallow;        # Return if everything allowed
foreach $test (@$disallow) {    # Lookup robots.txt
  if (index($object, $test) == 0) {
    $url->{'Error-Message'} = $url->{'Response'}->{'Error-Message'} =
           "Forbidden by <a href=\"$robots\" target=\"TEST\">robots.txt</a>";
    last;
  }
}
}

# Perform HTTP, requires method and URL, writes response to "Loaded"
# ==================================================================

sub Http ($$$) {
my ($this, $method, $url) = @_;
my %request;
$request{'method'} = $method;
$request{'url'} = $url;
$request{'file'} = $this->{'Temporary-File'}; # may always be supplied
if ($this->{'Version'}) {
  $request{'version'} = $this->{'Version'};
}
if ($this->{'Proxy'}) {
  $request{'proxy'} = $this->{'Proxy'};
}
my %headers = %{$this->{'User-Headers'}};
$headers{'User-Agent'} = $this->{'User-Agent'};
$headers{'Connection'} = 'close' unless $this->{'Keep-Alive'};
$headers{'Referer'} = $this->{'Referer'} if $this->{'Referer'};
delete $this->{'Referer'} unless $method eq 'HEAD';
$request{'headers'} = \%headers;
my %url;
$url{'URL'} = $url;
$url{'Method'} = $method;
$this->CheckRobots(\%url);
$this->{'Loaded'}->{$url} = \%url;
return \%url if exists $url{'Error-Message'};
my $response = $this->{'http'}->http(\%request);
if ($response && $response->{'Status'} == 403 && exists $request{'proxy'}) {
  delete $request{'proxy'};
  $response = $this->{'http'}->http(\%request);
}
if ($response) {
  $url{'Response'} = $response;
  $url{'Error-Message'} = $response->{'Error-Message'}
      if exists $response->{'Error-Message'};
  $url{'Status'} = $response->{'Status'} if exists $response->{'Status'};
  $url{'Type'} = lc($response->{'Content-Type'})
      if exists $response->{'Content-Type'};
}
else {
  $url{'Error-Message'} = 'No response';
}
$this->{'Loaded'}->{$url} = \%url;
return \%url;
}

# Load specified URL with HEAD
# ============================

sub Head ($$) {
my ($this, $url) = @_;
my $answer = $this->{'Loaded'}->{$url};
if (!$answer) {
  $answer = $this->Http('HEAD', $url);
}
return $answer;
}

# Parse HTML, gets \%url
# ======================

sub Parse ($$) {
my ($this, $url) = @_;
my $type = $url->{'Type'};
return unless $type =~ /^text\/html/; # Only HTML may be parsed
my $response = $url->{'Response'};
my $file = $response->{'file'};
return unless $file; # The file must exist...
open(HTML, $file);
my $parser = new IceBearSoft::ZWsgml(\*HTML);
my ($k, $v);
my (%body, %name, %href, %frame, %form, %meta);
$url->{'Body'} = \%body;
$body{'Name'} = \%name;
$body{'Href'} = \%href;
$body{'Frame'} = \%frame;
$body{'Form'} = \%form;
$body{'Meta'} = \%meta;
do {
  ($k, $v) = $parser->nextTag;
  if (length($k) > 0) {
    my %a;
    my $line = $parser->{'linenum'};
    if (length($v) > 0 && substr($k,0,1) ne '!') {
      %a = attributes($v);
    }
    if ($k eq 'base') {
      $body{'Base'} = $a{'href'};
    }
    elsif ($k eq 'form') {
      $form{$a{'action'}} = {'method' => $a{'method'}, 'line' => $line};
    }
    elsif ($k eq 'img') {
      $href{$a{'src'}} = {'type' => 'IMG', 'line' => $line};
    }
    elsif ($k eq 'body') {
      $href{$a{'background'}} = {'type' => 'BODY', 'line' => $line} if exists $a{'background'};
    }
    elsif ($k eq 'frame') { # this will always be loaded
      $frame{$a{'src'}} = {'target' => $a{'name'}, 'line' => $line};
    }
    elsif ($k eq 'a') {
      if (exists $a{'href'}) { # do not try to load "name"
        $href{$a{'href'}} = {'type' => 'A', 'line' => $line} unless exists $href{$a{'href'}};
      }
      if (exists $a{'name'}) { # label must start with hash
        my $label = '#' . $a{'name'};
        $name{$label} = $line;
      }
    }
    elsif ($k eq 'meta') {
      if ($a{'name'} =~ /^robots$/i) {
        $meta{'robots'} = {'content' => lc($a{'content'}), 'line' => $line};
      }
      if ($a{'http-equiv'} =~ /^refresh$/i) {
        if ($a{'content'} =~ /url=/is) {
          $meta{'refresh'} = {'content' => $', 'line' => $line};
        }
      }
    }
  }
} while (length($k) + length($v) > 0);
close HTML;  unlink $file;  delete $response->{'file'};
}

# Load specified URL with GET and parse it if Status == 200
# =========================================================

sub Get ($$) {
my ($this, $url) = @_;
my $answer = $this->{'Loaded'}->{$url};
my ($Status, $MethodUsed);
if ($answer) {
  $MethodUsed = $answer->{'Method'};
  $Status = $answer->{'Status'};
}
if ($MethodUsed ne 'GET' && (!$answer || $Status =~ /200|406/)) {
  my ($old_proxy, $response);
  $old_proxy = $this->{'Proxy'};
  if ($answer) { # Use the same proxy as was used in HEAD. This solves possible 305 msg
    $response = $answer->{'Response'};
    $this->{'Proxy'} = $response->{'proxy'} if $response;
  }
  $answer = $this->Http('GET', $url);
  $this->{'Proxy'} = $old_proxy;
  if ($answer->{'Status'} =~ /200|406/) {
    $this->Parse($answer);
  }
}
return $answer;
}

# Print the response headers
# ==========================

sub print_response ($$$) {
my ($this, $response, $id) = @_;
my $status = $response->{'Status'};
if ($this->{'Debug-Response'}) {
  print "<P><B>Response headers:</B>\n" or die "Browser died";
  print_ref $response;
  print "\n" or die "Browser died";
}
else {
  print "<table frame=border rules=rows border=5 cellpadding=2><tr>\n";
  print "<th align=right>Status:<td align=left>";
  print '<font color="#FF0000">' unless $status == 200;
  print "$status $response->{'Reason-Phrase'}";
  print '</font>' unless $status == 200;
  print "\n";
  print "<tr><th align=right>URL:<td align=left>$response->{'url'}\n";
  print "<tr><th align=right>Content-Type:<td align=left>$response->{'Content-Type'}\n";
  print "<tr><th align=right>Last-Modified:<td align=left>$response->{'Last-Modified'}\n";
  if ($status =~ /^30[1235]$/) {
    my $loctext = $status == 305 ? 'Use proxy:' : 'Location:';
    print "<tr><th align=right>$loctext<td align=left>$response->{'Location'}\n";
  }
  print "</table>\n" or die "Browser died";
}
if ($status >= 400 && $status != 406) {
  print "<P><font color=\"#FF0000\">Error-Message:</font> $status $response->{'Reason-Phrase'}\n";
}
if ($status >= 300) {
  $this->SummarizeError($status, $id);
}}

# ==========================================================================
# Checks existence of the supplied URL
# Accepts: ID, URL, Referer, Type of link (A, IMG, FRAME, BODY), line number
# Returns: nothing or new location
# ==========================================================================

sub CheckExists ($$$$$$) {
my ($this, $ID, $ch_url, $referer, $LinkType, $line) = @_;
my $XID = "<a name=\"$ID\">$ID</a>";
my $abs_url = abs_url $referer, $ch_url;
my $tag = $LinkType eq 'A' ? 'HREF' : $LinkType eq 'BODY' ? 'BACKGROUND' : 'SRC';
delete $this->{'Latest'};
$tag = "&lt;$LinkType $tag=&quot;";
my $lninfo = $line > 0 ? ", line #$line" : ' (by redirection)';
print "<H3>$XID $tag<A TARGET=\"TEST\" HREF=\"$abs_url\">$ch_url</A>&quot;&gt;</H3>\n";
print "<P>Referred from <A TARGET=\"TEST\" HREF=\"$referer\">$referer</A>$lninfo<P>\n"
      or die "Browser died";
my %h = split_url $abs_url;
if ($h{'scheme'} ne 'http') {
  print "<P><font color=\"#FF0000\">Warning: Verification of this kind of URL is not supported.</font>\n"
        or die "Browser died";
  $this->SummarizeError('Unsupported URL type (' . $h{'scheme'} . ':)', $ID);
  return;
}
my $label = $h{'label'}; # this may be undefined
delete $h{'label'};
my $url2test = merge_url %h;
my $url;
$this->{'Referer'} = $referer;
if ($label) {
  $url = $this->Get($url2test);
}
else {
  $url = $this->Head($url2test);
}
delete $this->{'Referer'}; # if it was not done elsewhere...
# Get the status
my $status;
if ($url) {
  $status = $url->{'Status'};
}
if ($status =~ /200|406/ && !$label && $url->{'Type'} =~ /^text\/html/i) {
  $this->{'Referer'} = $referer;
  my $g = $this->Get($url2test); # Get and parse the page
  delete $this->{'Referer'};
  $url = $g if $g && $g->{'Status'} eq 200 && $g->{'Body'};
}
$this->{'Latest'} = $url;
if (!$status) { # not loaded, get Error-Message
  print "<P>The resource was not loaded.\n" or die "Browser died.";
  if ($url) {
    print "<BR><font color=\"#FF0000\">Error-Message:</font> $url->{'Error-Message'}\n"
    or die "Browser died.";
  }
  my $errtype = $url->{'Error-Message'} =~ />robots.txt</ ?
     'Forbidden by robots.txt' : $url->{'Error-Message'};
  $this->SummarizeError($errtype, $ID);
  return;
}
my $response = $url->{'Response'};
$this->print_response($response, $ID);
if (($LinkType =~ /(IMG|BODY)/) && $status eq 200 && $url->{'Type'} !~ /^image\//i) {
  print "<P><font color=\"#FF0000\">Error-Message: Wrong type of image object</font>\n"
        or die "Browser died";
  $this->SummarizeError('Wrong type of image object', $ID);
}
if ($status == 305) {
  my $old_proxy = $this->{'Proxy'}; # may be undefined
  if ($old_proxy eq $response->{'Location'}) {
    print "<P><font color=\"#FF0000\">Error-Message: The proxy has been used but the server complains, ";
    print "contact the administrator!</font>\n" or die "Browser died";
    $this->SummarizeError('Bad proxy', $ID);
    return; # Sorry, there is no help
  }
  print "<P>Trying again via the specified proxy.\n" or die "Browser died";
  delete $this->{'Loaded'}->{$url2test}; # otherwise we get the same 305
  $this->{'Proxy'} = $response->{'Location'};
  my $retvalue = $this->CheckExists($ID.'*', $ch_url, $referer, $LinkType, $line);
  delete $this->{'Proxy'};
  $this->{'Proxy'} = $old_proxy if $old_proxy; # restore the proxy
  return $retvalue;
}
if ($status == 200) {
  if ($label) {
    my $lbl = '#'.$label;
    my $body = $url->{'Body'};
    if (!$body || !exists $body->{'Name'}->{$lbl}) {
      print "<P><font color=\"#FF0000\">Label $lbl does not exist.</font>\n"
            or die "Browser died";
      $this->SummarizeError('Label does not exist', $ID);
    }
  }
  if (!exists($url->{'Checked'}) && $url->{'Type'} =~ /^text\/html/) {
    if ($h{'host'} eq $this->{'This'}) {
      my $loc = $this->{'Local-URL'};
      push (@$loc, $url2test); # push local URL
      $this->{'Local-URL'} = $loc; # return it back, otherwise it is forgotten
      $url->{'Checked'} = 'PUSHED'; # avoid pushing it again
      $this->{'Loaded'}->{$url2test} = $url; # return it to the hash
    }
  }
}
if ($status =~ /^30[1-3]$/) {
  return abs_url $abs_url, $response->{'Location'}; # return new location
}
# Return with an undefined value
}

# Checks <META refresh> of the object stored in $this->{'Latest'}, returns URL or nothing
# =======================================================================================

sub CheckRefresh ($$) {
my ($this, $id) = @_;
my $url = $this->{'Latest'};
return unless $url;
delete $this->{'Latest'};
my $body = $url->{'Body'};
return unless $body;
my $meta = $body->{'Meta'};
return unless $meta;
my $refresh = $meta->{'refresh'};
return unless $refresh;
print '<P><font color="#FF0000"><code>&lt;META&gt;</code> redirection: </font>';
print "$refresh->{'content'}\n" or die "Browser died";
$this->SummarizeError('META redirection', $id);
return (abs_url($url->{'URL'}, $refresh->{'content'}), $refresh->{'line'});
}

# LinkCheck -- the main function, requires URL and returns the list of local URLs
# ===============================================================================

sub Linkcheck ($$) {
my ($this, $url) = @_;
$this->{'Local-URL'} = (); # Make an empty array
if ($this->{'Follow'} eq 'Incremental') {
  $this->PushLinks($url);
}
my @urls = ($url); # list of URLs to check
my $uid = 1 + $this->{'ID'};
my $redircount = 0;
my $helpDisplayed = 0;
while ($#urls >= 0) {
  my $xuid = "<a name=\"$uid\">$uid</a>";
  $url = shift @urls;
  $helpDisplayed = 1;
  my %hr = split_url $url;
  $this->{'This'} = $hr{'host'} unless $this->{'This'};
  if ($hr{'scheme'} ne 'http') {
    $helpDisplayed = 0;
    print "<H2>$xuid <A TARGET=\"TEST\" HREF=\"$url\">$url</A></H2>\n";
    print "<P><font color=\"#FF0000\">Warning: Verification of this kind of URL is not supported.</font>\n"
          or die "Browser died";
    $this->SummarizeError('Unsupported URL type (' . $hr{'scheme'} . ':)', $ID);
    next;
  }
  my $object = $this->Get($url);
  if (!$object) { # There was no response
    print "<H2>$xuid <A TARGET=\"TEST\" HREF=\"$url\">$url</A></H2>\n";
    print "<P><font color=\"#FF0000\">Error-Message: Sorry, there was no response.</font>\n"
          or die "Browser died";
    next;
  }
  next if $object->{'Checked'} eq 'YES';
  $object->{'Checked'} = 'YES';
  $this->{'Loaded'}->{$url} = $object; # otherwise it may be forgotten
  $helpDisplayed = 0;
  print "<H2>$xuid <A TARGET=\"TEST\" HREF=\"$url\">$url</A></H2>\n";
  my $status = $object->{'Status'};
  if (!$status) {
    print "<P>The resource was not loaded.\n";
    print "<BR><font color=\"#FF0000\">Error-Message:</font> $object->{'Error-Message'}\n"
          or die "Browser died";
    my $errtype = $url->{'Error-Message'} =~ />robots.txt</ ?
       'Forbidden by robots.txt' : $url->{'Error-Message'};
    $this->SummarizeError($errtype, $ID);
    next;
  }
  my $response = $object->{'Response'};
  $this->print_response($response, $uid);
  if ($status == 305) {
    my $old_proxy = $this->{'Proxy'}; # may be undefined
    if ($old_proxy eq $response->{'Location'}) {
      print "<P><font color=\"#FF0000\">Error-Message: The proxy has been used but the server complains, ";
      print "contact the administrator!</font>\n"
            or die "Browser died";
      $this->SummarizeError('Bad proxy', $uid);
      next; # Sorry, there is no help
    }
    print "<P>Trying again via the specified proxy.\n" or die 'Browser died';
    delete $this->{'Loaded'}->{$url}; # otherwise we get the same 305
    $this->{'Proxy'} = $response->{'Location'};
    $object = $this->Get($url);
    delete $this->{'Proxy'};
    $this->{'Proxy'} = $old_proxy if $old_proxy; # restore the proxy
    $status = $object->{'Status'};
    if (!$status) {
      print "<P>The resource was not loaded.\n";
      print "<BR><font color=\"#FF0000\">Error-Message:</font> $object->{'Error-Message'}\n"
            or die 'Browser died';
      my $errtype = $url->{'Error-Message'} =~ />robots.txt</ ?
         'Forbidden by robots.txt' : $url->{'Error-Message'};
      $this->SummarizeError($errtype, $ID);
      next;
    }
    $response = $object->{'Response'};
    $this->print_response($response, $uid);
  }
  if ($status =~ /^30[1-3]$/) { # Check the new location
    $redircount++;
    if ($redircount < 5) {
      unshift @urls, $response->{'Location'};
      next;
    }
  }
  $this->RedirError($uid, $response->{'Location'}) unless $redircount < 5;
  next if $status !~ /200|406/; # not loaded
  unless ($response->{'Content-Type'} =~ /^text\/html/i) {
    print "<P><font color=\"#FF0000\">Only HTML files can be checked.</font>\n"
          or die 'Browser died';
    next;
  }
  my $linkid = 0;
  my $body = $object->{'Body'};
  if (!$body) {
    print "<P><font color=\"#FF0000\">The resource body was not loaded.</font>\n"
          or die 'Browser died';
      my $errtype = $url->{'Error-Message'} =~ />robots.txt</ ?
         'Forbidden by robots.txt' : $url->{'Error-Message'};
      $this->SummarizeError($errtype, $ID);
    next;
  }
  # META handling
   if (exists $body->{'Meta'}) {
     my $meta = $body->{'Meta'};
     if ($meta->{'robots'} =~ /\bnofollow\b/is) {
       print '<P><font color="#FF0000">The <code>&lt;META&gt;</code> element does not ';
       print "allow robots to follow the links within this page.</font>\n"
             or die 'Browser died';
       $this->SummarizeError('META nofollow', $uid);
       next;
     }
     if (exists $meta->{'refresh'}) {
       my $refresh = $meta->{'refresh'};
       $redircount++;
       print '<P><font color="#FF0000"><code>&lt;META&gt;</code> redirection: </font>';
       print "$refresh->{'content'}\n" or die 'Browser died';
       $this->SummarizeError('META refresh', $uid);
       $this->RedirError($uid, $refresh->{'content'}) unless $redircount < 5;
       unshift @urls, abs_url $url, $refresh->{'content'} if $redircount < 5;
     }
     else {
       $redircount = 0;
     }
   }
   else {
     $redircount = 0;
   }
  my $referer = $body->{'Base'} ? $body->{'Base'} : $url;
  # Warn that forms are not checked
  my $form = $body->{'Form'};
  if (scalar(%$form) > 0) {
    print "<P><font color=\"#FF0000\">Warning: Verification of the following form(s) is not supported:</font>\n";
    print_ref $form;
    $this->SummarizeError('Form(s) not verified', $uid);
  }
  # Display the link to the help window
  $this->HelpWindow();  $helpDisplayed = 1;
  # Checks images and references
  my $href = $body->{'Href'};
  my $frame = $body->{'Frame'};
  my $alllinks = scalar(keys %$href) + scalar(keys %$frame);
  foreach $url (keys %$href) {
    my $LinkType = $href->{$url}->{'type'};
    my $line = $href->{$url}->{'line'};
    my $redircount = 0;
    my $Xurl;
    while ($url) {
      $linkid++;  $redircount++;  $Xurl = $url;
print<<"JS";
<script language="JavaScript">
<!--
putSubtitle($linkid, $alllinks);
// -->
</script>
JS
      $url = $redircount < 5 ?
             $this->CheckExists("$uid.$linkid", $url, $referer, $LinkType, $line)
             : '';
      ($url, $line) = $this->CheckRefresh("$uid.$linkid") if !$url && $redircount < 5;
      $alllinks++ if $url;
    }
    $this->RedirError("$uid.$linkid", $Xurl) unless $redircount < 5;
  } continue {
    $this->HelpWindow();
  }
  # Check frames
  foreach $url (keys %$frame) {
    my $line = $frame->{$url}->{'line'};
    my $redircount = 0;
    while ($url) {
      my $frm = $url;
      $linkid++;  $redircount++;
print<<"JS";
<script language="JavaScript">
<!--
putSubtitle($linkid, $alllinks);
// -->
</script>
JS
      $url = $redircount < 5 ?
             $this->CheckExists("$uid.$linkid", $url, $referer, 'FRAME', $line)
             : '';
      ($url, $line) = $this->CheckRefresh("$uid.$linkid") if !$url && $redircount < 5;
      unshift @urls, abs_url($referer, $frm)
              unless $url and $redircount >= 5; # Frames must be checked
      $alllinks++ if $url;
    }
    $this->RedirError("$uid.$linkid", $url) unless $redircount < 5;
  } continue {
    $this->HelpWindow();
  }
} continue {
  $uid++;
  $this->HelpWindow() unless $helpDisplayed;
}
$this->{'ID'} = $uid - 1;
my $LocalURLs = $this->{'Local-URL'};
return @$LocalURLs;
}

# This function parses information about a previously loaded object
# =================================================================

sub ParseLoaded ($$) {
my ($this, $info) = @_;
if ($info =~
 /^\(([^)]*)\)\(([^)]*)\)\(([^)]*)\)\(([^)]*)\)\(([^)]*)\)\(([^)]*)\)\(([^)]*)\)/) {
#        1          2          3          4          5          6          7
  my $urlinfo = $1;
  my $responseinfo = $2;
  my $forminfo = $3;
  my $frameinfo = $4;
  my $hrefinfo = $5;
  my $metainfo = $6;
  my $nameinfo = $7;
  my $reason = $';
  my (%object,%body,%response,%form,%frame,%href,%meta,%name);
#                       1       2       3       4       5       6
  if ($urlinfo =~ /^([^#]*)#([^#]*)#([^#]*)#([^#]*)#([^#]*)#([^#]*)$/) {
    my $url = $1;
    $object{'Previous'} = 'YES';
    $response{'Previous'} = 'YES';
    $object{'URL'} = $url;
    $object{'Checked'} = $2 if $2;
    $object{'Error-Message'} = $3 if $3;
    $object{'Method'} = $4 if $4;
    $object{'Status'} = $5 if $5;
    $object{'Type'} = $6 if $6;
    $response{'url'} = $url;
    $response{'Error-Message'} = $3 if $3;
    $response{'Request-Method'} = $4 if $4;
    $response{'Status'} = $5 if $5;
    $response{'Content-Type'} = $6 if $6;
    $response{'Reason-Phrase'} = $reason if $reason;
#                              1       2       3
    if ($responseinfo =~ /^([^#]*)#([^#]*)#([^#]*)$/) {
      $response{'proxy'} = $1 if $1;
      $response{'Last-Modified'} = $2 if $2;
      $response{'Location'} = $3 if $3;
    }
    $object{'Response'} = \%response;
    if ($object{'Method'} eq 'GET') {
      if ($forminfo) {
        while ($forminfo =~ /^(.*?);;/) {
          my $keyval = $1;  $forminfo = $';
          if ($keyval =~ /::/) {
            $form{$`} = $';
          }
        }
      }
      if ($forminfo) {
        while ($forminfo =~ /^(.*?);;/) {
          my $keyval = $1;  $forminfo = $';
          if ($keyval =~ /::/) {
            $form{$`} = $';
          }
        }
      }
      if ($frameinfo) {
        while ($frameinfo =~ /^(.*?);;/) {
          my $keyval = $1;  $frameinfo = $';
          if ($keyval =~ /::/) {
            $frame{$`} = $';
          }
        }
      }
      if ($hrefinfo) {
        while ($hrefinfo =~ /^(.*?);;/) {
          my $keyval = $1;  $hrefinfo = $';
          if ($keyval =~ /::/) {
            $href{$`} = $';
          }
        }
      }
      if ($metainfo) {
        while ($metainfo =~ /^(.*?);;/) {
          my $keyval = $1;  $metainfo = $';
          if ($keyval =~ /::/) {
            $meta{$`} = $';
          }
        }
      }
      if ($nameinfo) {
        while ($nameinfo =~ /^(.*?);;/) {
          my $val = $1;  $nameinfo = $';
          $name{$1} = 0;
        }
      }
      $body{'Name'} = \%name;
      $body{'Href'} = \%href;
      $body{'Frame'} = \%frame;
      $body{'Form'} = \%form;
      $body{'Meta'} = \%meta;
      $object{'Body'} = \%body;
    }
    $this->{'Loaded'}->{$url} = \%object;
  }
}}

# This function inserts the objects which were previously loaded
# ==============================================================

sub InsertLoaded ($$) {
my ($this, $cgivars) = @_;
if (ref $cgivars->{'Loaded'}) {
  my $arr = $cgivars->{'Loaded'};
  foreach $info (@$arr) {
    $this->ParseLoaded($info);
  }
}
else {
  my $info = $cgivars->{'Loaded'};
  $this->ParseLoaded($info);
}
}

# Export hash (only scalar elements)
# ==================================

sub ExportHash ($$) {
my ($this, $hash) = @_;
my $retvalue = '(';
foreach $key (keys %$hash) {
  if (!ref($hash->{$key})) {
    $retvalue .= $key.'::'.$hash->{$key}.';;';
  }
}
$retvalue .= ')';
return $retvalue;
}

# Export keys from a hash
# =======================

sub ExportKeys ($$) {
my ($this, $hash) = @_;
my $retvalue = '(';
foreach $key (keys %$hash) {
  $retvalue .= $key.';;';
}
$retvalue .= ')';
return $retvalue;
}

# Export a single object
# ======================

sub ExportObject ($$) {
my ($this, $object) = @_;
my ($main, $respinfo, $forminfo, $frameinfo, $hrefinfo, $metainfo, $nameinfo, $reason);
my $delim = '(';
for $k ('URL','Checked','Error-Message','Method','Status','Type') {
  $main .= $delim.$object->{$k};  $delim = '#';
}
$main .= ')';
$delim = '(';
my $response = $object->{'Response'};
if ($response) {
  $reason = $response->{'Reason-Phrase'};
  for $k ('proxy', 'Last-Modified', 'Location') {
    $respinfo .= $delim.$response->{$k};  $delim = '#';
  }
  $respinfo .= ')';
  my $body = 0;
  if ($object->{'Method'} eq 'GET') {
    $body = $object->{'Body'};
  }
  if ($body) {
    if ($object->{'Checked'} eq 'YES') {
      # only meta and names are needed
      $forminfo = $frameinfo = $hrefinfo = '()';
      $metainfo = $this->ExportHash($body->{'Meta'});
      $nameinfo = $this->ExportKeys($body->{'Name'});
    }
    else {
      # export everything
      $forminfo = $this->ExportHash($body->{'Form'});
      $frameinfo = $this->ExportHash($body->{'Frame'});
      $hrefinfo = $this->ExportHash($body->{'Href'});
      $metainfo = $this->ExportHash($body->{'Meta'});
      $nameinfo = $this->ExportKeys($body->{'Name'});
    }
  }
  else {
    # no body
    $forminfo = $frameinfo = $hrefinfo = $metainfo = $nameinfo = '()';
  }
}
else {
  # no response
  $respinfo = $forminfo = $frameinfo = $hrefinfo = $metainfo = $nameinfo = '()';
}
my $value = $main.$respinfo.$forminfo.$frameinfo.$hrefinfo.$metainfo.$nameinfo.$reason;
print "<input type=hidden name=\"Loaded\" value=\"$value\">\n" or die 'Browser died';
}

# Export loaded objects
# =====================

sub ExportLinks ($) {
my $this = shift;
my $loaded = $this->{'Loaded'};
foreach $k (keys %$loaded) {
  $this->ExportObject($loaded->{$k});
}}

# Push local links which are marked as pushed, do not push the base url
# =====================================================================

sub PushLinks ($$) {
my ($this, $BaseURL) = @_;
my $loaded = $this->{'Loaded'};
foreach $url (keys %$loaded) {
  next if $url eq $BaseURL;
  my $object = $loaded->{$url};
  if ($object->{'Checked'} eq 'PUSHED') {
    my $loc = $this->{'Local-URL'};
    push (@$loc, $url); # push local URL
    $this->{'Local-URL'} = $loc; # return it back, otherwise it is forgotten
  }
}}

# This function starts the progcessing
# ====================================

sub go ($) {
my $this = shift;
$| = 1;
my $method = uc $ENV{'REQUEST_METHOD'}; # Check method, only POST is allowed
if ($method ne 'POST') {
print<<"EOD";
Status: 405
Content-Type: text/html

<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">
<HTML>
 <HEAD>
  <TITLE>405 Method not allowed</TITLE>
 </HEAD>
<BODY>
  <H1>405 Method not allowed</H1>
  <P>Request method <code>$method</code> is not allowed for this object.
</BODY>
</HTML>

EOD

die "Method $method not allowed";
}

# Prepare heading

print<<"EOD";
Content-Type: text/html

<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2 Final//EN\">
EOD

my $helpwin = $this->{'LinkChecker-Help-Window'};
print<<"EOD";
<HTML>
 <HEAD>
  <TITLE>Results of $this->{'User-Agent'}</TITLE>

<script language=\"JavaScript\">
<!--
var basetitle = 'Testing #1';
function putTitle(t, r) {
basetitle = 'Testing #' + t + '; remaining ' + r;
window.document.title = basetitle;
}
function putSubtitle(l, a) {
window.document.title = basetitle + ' (' + l + '/' + a + ')';
}
// -->
</script>

 </HEAD>
 <BODY
  BGCOLOR=\"#FFFFFF\"
  TEXT=\"#000000\"
  LINK=\"#0000FF\"
  VLINK=\"#0000FF\"
  ALINK=\"#FF0000\"
 >
<H1>Results of $this->{'User-Agent'}</H1>
<P>This is the result of the link check. All URLs are activated and may be open
in a separate window. If you have Mozilla 1.0 or later, it may be more comfortable to Ctrl-click
the URL which will open it in its own tab instead of a separate window (but this assumes that this
function is enabled in your browser).
<P>If you do not uderstand any message, you can open
<a target=\"HELP\" href=\"$helpwin\">help</a>
in its own window.

<P>After processing is finished, <a href=\"#summary\">summary table</a> will be displayed (the
function is not yet fully debugged).

<script language="JavaScript">
<!--
var now = new Date();
document.writeln('<P><font color="green">Verification started at ' + now.toLocaleString() + '</font>');
// -->
</script>
EOD

# Get the form contents
my %cgivars = cgi_parse;
if ($cgivars{'KeepAlive'} eq 'on') {
  $this->{'Keep-Alive'} = 1;
}
if ($cgivars{'1.0'} eq 'on') {
  $this->{'Version'} = '1.0';
}
if ($cgivars{'Debug'} eq 'on') {
  $this->{'Debug-Response'} = 1;
}
if ($cgivars{'Proxy'}) {
  $this->{'Proxy'} = $cgivars{'Proxy'};
}
if ($cgivars{'UseProxy'}) {
  $this->{'Proxy'} = $ENV{'DEFAULT_PROXY'};
}
if (exists($cgivars{'ID'})) {
  $this->{'ID'} = $cgivars{'ID'};
}
$this->{'Follow'} = $cgivars{'Follow'};
if ($cgivars{'AcceptLanguage'} =~ /^\s*(\S.*?)\s*$/) {
  $this->{'User-Headers'}->{'Accept-Language'} = $1;
}
if ($cgivars{'AcceptCharset'} =~ /^\s*(\S.*?)\s*$/) {
  $this->{'User-Headers'}->{'Accept-Charset'} = $1;
}
for ($i = 1;  $i <= 3;  $i++) {
  if ($cgivars{"hdrname$i"} =~ /^\s*(\S.*?)\s*$/) {
    my $hdrname = $1;
    if ($cgivars{"hdrvalue$i"} =~ /^\s*(\S.*?)\s*$/) {
      $this->{'User-Headers'}->{$hdrname} = $1;
    }
  }
}
$this->{'debug'} = $cgivars{'Show'};

if ($cgivars{'Follow'} eq 'Incremental') {
  $this->InsertLoaded(\%cgivars);
}

# Check the base URL
my $base = http_url($cgivars{'URL'});
my @LocalURLs = $this->Linkcheck($base);

# Check other references
if (($cgivars{'Follow'} eq 'Local') || ($cgivars{'Follow'} eq 'Below')) {
  while ($#LocalURLs >= 0) {
my $pageno = 1 + $this->{'ID'};
print<<"JS";
<script language=\"JavaScript\">
<!--
putTitle($pageno, $#LocalURLs);
// -->
</script>
JS
    $url = shift @LocalURLs;
    if ($cgivars{'Follow'} eq 'Below') {
      next unless $url =~ /^$base/;
    }
    push @LocalURLs, $this->Linkcheck($url);
  }
}
print<<"JS";
<script language=\"JavaScript\">
<!--
window.document.title='Preparing the summary table';
// -->
</script>
JS

delete $this->{'Proxy'} if $cgivars{'UseProxy'};

# Summary table
print<<'EOD';
<H1><a name="summary">Errors and warnings summary</a></H1>
EOD

my $summary = $this->{'Summary'};
my @errcodes = keys %$summary;

if ($#errcodes < 0) {
  print "<P>No errors, no warnings.\n";
} else {
  print<<EOD;
<P><table frame=box rules=all border=5 cellpadding=2><tr>
<th>Error type<th>Count<th>Last occurence
EOD
  foreach (sort @errcodes) {
    my $err = $summary->{$_};
    my $cnt = $err->{'count'};
    my $occ = $err->{'last'};
    print "<tr><td>$_<td align=right>$cnt<td align=right><a href=\"#$occ\">$occ</a>\n";
  }
  print "</table>\n";
}

print<<"JS";
<script language=\"JavaScript\">
<!--
window.document.title='$this->{'User-Agent'}: verification finished.';
var then = new Date();
document.writeln('<P><font color="green">Verification finished at ' + then.toLocaleString() + '</font>');
// -->
</script>
JS

if ($ENV{'HTTP_REFERER'}) {
  my $lcf = $ENV{'HTTP_REFERER'};
  print "<P><HR><P><a href=\"$lcf\">New LinkChecker form</a>\n";
}

# Show local variables (if requested)
if ($cgivars{'Show'} eq 'on') {
  my %hash = %$this;
  print "<P><HR><B>";
  print_ref $this;
  print "</B>\n";
  print_ref \%hash;
}

# Incremental checking (output)
if ($#LocalURLs >= 0 && ($cgivars{'Follow'} eq 'Incremental')) {
  my $action = $ENV{'REQUEST_URI'};
print<<"EOD";
<P><HR>
<form method=\"POST\" action=\"$action\">
<P>URL:
<select height=1 name=\"URL\">
EOD

  foreach $u (sort @LocalURLs) {
    print "<option>$u</option>\n";
  }
  print "</select>\n";
  print "<input type=submit value=\"Check it!\">\n<P>";
  $cgivars{'ID'} = $this->{'ID'};
  for $k ('KeepAlive','1.0','Debug', 'Show','Follow','UseProxy','Proxy', 'ID') {
    if ($cgivars{$k}) {
      my $v = url_encode $cgivars{$k};
      print "<input type=hidden name=\"$k\" value=\"$v\">\n";
    }
  }
  my $hname = 'hdrname0';
  my $hval = 'hdrvalue0';
  for $k (keys %{$this->{'User-Headers'}}) {
    my $v = url_encode $this->{'User-Headers'}->{$k};
    if ($k =~ /^(Accept)-(Language|Charset)$/) {
      $k = "$1$2";
    }
    else {
      $hname++;  $hval++;
      my $w = url_encode $k;
      print "<input type=hidden name=\"$hname\" value=\"$w\">\n";
      $k = $hval;
    }
    print "<input type=hidden name=\"$k\" value=\"$v\">\n";
  }
  $this->ExportLinks;
  print "</form>\n";
}

# Trailer
print<<"EOD";
<P><HR>
$this->{'User-Agent'} was developed by<i>
<A href=\"http://icebearsoft.euweb.cz/\">Z. Wagner, Ice Bear Soft</A></i>.
</BODY>
</HTML>
EOD
}

# Destruction (must close connection if kept alive and delete the temporary file)

sub DESTROY {
my $this = shift;
delete $this->{'http'}; # This should be sufficient.
unlink $this->{'Temporary-File'};
}
