package IceBearSoft::Zwebfun;

# World Wide Web function(s)
# (C) Z. Wagner -- Ice Bear Soft, 4 Aug 1998
# Last update: 1 Dec 1998 (Persistent connection)
# 28 Dec 1998 (Trasfer-Encoding: chunked)
# 29 Jan 1999 (Changes in timeout, forcing version 1.0 if server responded that way)
#  9 Jun 1999 (Parameters preferably sent as HASHREF)
# 11 Mar 2000 (Entity field name normalized to ETag instead of implied Etag)
# 11 Sep 2001 (Added User-Info)
# 16 Jan 2002 (Added mkpath befor opening the file)
# 31 Jul 2003 (User-Info not properly handled - $request [asked from %request])
# 13 Aug 2003 (Host header added to the HEAD request when closing the connection)

use Carp;
use Socket;
use File::Path;
require Exporter;
@ISA = qw(Exporter);

sub new($*);

{
chop($hostname = lc(`hostname`));
$crlf = "\015\012";  $lf = "\012";
$default_proxy_timeout = 10; # timeout to wait for proxy verification
$wrong_chars = '[\\\* \000-\037<>"|\177-\377]'; # not allowed in URL (I hope...)
$retry = 3;
$sockaddr = 'S n a4 x8';
$default_lifetime = 600; # maximum lifetime
$default_version = 1.1; # default http version
}

# Create a new web function object, no arguments

sub new {
my $class = shift;
my $this = {};
my $hosts = {};
my $versions = {};
$this->{'hosts'} = $hosts;
$this->{'versions'} = $versions;
my ($name, $aliases, $proto) = getprotobyname('tcp');
$this->{'proto'} = $proto;
my ($type, $len, $thisaddr);
($name, $aliases, $type, $len, $thisaddr) = gethostbyname($hostname);
$this->{'hosts'}->{$hostname} = $thisaddr;
$this->{'home'} = pack($sockaddr, AF_INET, 0, $thisaddr);
$this->{'port'} = 80;
$this->{'read-timeout'} = 30;
return bless $this, $class;
}

# Read bytes, either the specified number or till the line feed

sub ReadBytes ($;$) {
my $this = shift;
my ($bytes, $len, $x, $o_alarm, $rcvd, $sz);
$len = $#_ >= 0 ? shift : 0;
$len = 0 if $len < 0;
$rcvd = $len > 0 ? 0 : -1;
my $timeout = $this->{'read-timeout'};
$bytes = '';
my $Socket = $this->{'socket'};
return $bytes unless $Socket;
if (time() > $this->{'lifetime'}) {
  close $Socket;
  $this->Clean();
  return $bytes;
}
if ($timeout > 0) {
  $o_alarm = $SIG{'ALRM'};
  $SIG{'ALRM'} = sub {die;};
}
while ($rcvd < $len) {
  my $OK = 1;
  alarm($timeout) if $timeout > 0;
  eval {
    $sz = read($Socket, $x, 1);
  };
  alarm(0);
  $OK = 0 if $@;
  if ($sz) {
    if ($len > 0) {
      $rcvd += $sz;
    }
    else {
      $rcvd = $len + 1 if $x eq $lf;
    }
    $bytes .= $x;
  }
  else {
    last;
  }
  if (!$OK) {
    close $Socket;
    $this->Clean();
    $rcvd = $len + 1;
  }
}
$SIG{'ALRM'} = $o_alarm if $timeout > 0;
return $bytes;
}

# Get the host address, accepts this and host, returns IP (in perl form)

sub host_address {
my ($this, $host) = @_;
my $ip;
if ($host =~ /^\s*((\d+\.){3}\d+)\s*$/) {
  $ip = pack('C4', split(/\./, $1));
}
else { # host lookup
  $host = lc($host);    # case insensitive search
  $ip = $this->{'hosts'}->{$host};
  if (!$ip) {
    $ip = (gethostbyname($host))[4];
    $this->{'hosts'}->{$host} = $ip if $ip;
  }
}
return $ip;
}

# Cleans the socket KeepAlive

sub Clean {
my $this = shift;
delete $this->{'socket'};
delete $this->{'sockaddr'};
}

# Close an open socket (must communicate with the server)

sub CloseConnection {
my $this = shift;
my $S = $this->{'socket'};
if ($S) {
my $oldfh = select($S);
my $h = $this->{'LastHost'};
print<<"EOD";
HEAD / HTTP/1.1
Host: $h
Connection: close

EOD
select($oldfh);
while($this->ReadBytes) {};
$S = $this->{'socket'}; # may be closed due to error
close $S if $S;
}
$this->Clean();
}

# Reconnects the same socket

sub Reconnect {
my ($this, $timeout) = @_;
my $address = $this->{'sockaddr'};
$timeout = 0 if !$timeout;
my $failed = $timeout > 0 ? 1 : 0;
my $old_alarm;
if ($timeout > 0) {
  $old_alarm = $SIG{'ALRM'}; # This may insert an undefined address.
  $SIG{'ALRM'} = sub { die };
  alarm($timeout);
}
eval {
    last unless socket(S, AF_INET, SOCK_STREAM, $this->{'proto'});
    last unless bind(S, $this->{'home'});
    last unless connect(S, $address);
    $failed = 0;
};
if ($timeout > 0) {
  alarm(0);
  $SIG{'ALRM'} = $old_alarm; # The variable may be uninitialized, see above.
}
if ($failed) {
  close(S);  return 0;
}
my $oldfh = select(S);  $| = 1;  select($oldfh);  binmode(S);
return $this->{'socket'} = \*S;
}

# Open, bind and connect to a socket
# Accepts this, address and optionally timeout
# Returns the socket handle (also stored in $this->{'socket'}

sub Connect {
my ($this, $address, $timeout) = @_;
if (exists $this->{'socket'}) {
  return $this->{'socket'} if $address == $this->{'sockaddr'};
  $this->CloseConnection();
}
$this->{'sockaddr'} = $address;
$sock = $this->Reconnect($timeout);
$this->Clean() if !$sock;
return $sock;
}


# Performs the http function, accepts \%request (or %request) containing:
#       method - only HEAD, GET, POST, OPTIONS, TRACE are supported
#       url
#       version - optional, default = 1.1
#       file - used for GET or POST only, default = STDOUT
#       data - URL encoded data for the POST method
#       proxy - optional proxy server
#       lifetime - maximum HTTP lifetime in seconds
#       headers - optional list of other headers, this may contain:
#               - single string with "key: value"
#               - ARRAYREF to an array of strings
#               - HASHREF to a hash of key/value pairs
#       User-Info - any information which will just be copied to the response
#       Content-Type - for POST only, default = application/x-www-form-urlencoded
# Returns \%response containing the headers, url and file are copied from %request.
# If the \%response does not contain Status, an error occured. See Error-Message.
# if the return value is undefined, the protocol is not supported or the socket died.

# For 206 the function accepts the range unless the size of the target file is shorter
# than the first byte.

sub http {
  my $this = shift;
  my $request;
  my %req;
  if ($#_ == 0 && ref($_[0]) eq 'HASH') {
    $request = shift;
  }
  else {
    %req = @_;  $request = \%req;
  }
  my %response;
  my $method = $request->{'method'};
  croak "Method '". $method. "' not allowed"
    unless $method =~ /^(GET|POST|HEAD|OPTIONS|TRACE)$/;
  if ($method eq 'POST') {
    croak 'No data to post' unless exists($request->{'data'});
    $request->{'Content-Type'} = 'application/x-www-form-urlencoded'
      unless exists($request->{'Content-Type'});
    $request->{'Content-Length'} = length($request->{'data'});
  }
  my ($protocol, $host, $port, $object, $proxy, $url);
  $port = $this->{'port'};  $object = '/';  #defaults
  if (($url = $request->{'url'}) =~ m!^([a-z]+)://([^:/\*]+)(:\d+)?(.*)$!) {
    $protocol = lc($1) if $1;
    $host = lc($2) if $2;
    $port = $3 if $3;
    $object = $4 if $4;
    $port =~ s/://;
  }
  else {
    $response{'Error-Message'} = "Invalid URL $request->{'url'}";
    return \%response;
  }
  return if $protocol ne 'http';
  if ($url =~ /$wrong_chars/) {
    $response{'Error-Message'} = "URL contains illegal characters";
    return \%response;
  }
  $this->{'lifetime'} = ($request->{'lifetime'} ? $request->{'lifetime'} : $default_lifetime)
                      + time();
  my $version = $request->{'version'} ? $request->{'version'} : $default_version;
  if (exists $request->{'proxy'}) {
    my $use_proxy = 1;
    if (exists $request->{'no-proxy'}) {
      foreach (@{$request->{'no-proxy'}}) {
        $use_proxy = 0 if $host =~ /$_/;
      }
    }
    if ($use_proxy) {
      $proxy = $request->{'proxy'};
      $proxy_port = 80;
      if ($proxy =~ /:/) {
        $proxy = $`;  $proxy_port = $';
      }
    }
  }

  # Copy the url and file into the %response

  $response{'url'} = $url;
  my $file = $request->{'file'};
  $response{'file'} = $file if $file;
  $response{'User-Info'} = $request->{'User-Info'} if exists $request->{'User-Info'};

  # Get the host address

  my $address = $this->host_address($proxy ? $proxy : $host);
  if ($proxy && $address) {
    my $proxy_timeout = $default_proxy_timeout;
    if (exists $request->{'proxy_timeout'}) {
      $proxy_timeout = $request->{'proxy_timeout'};
    }
    $where = pack($sockaddr, AF_INET, $proxy_port, $address);
    $Socket = $this->Connect($where,$proxy_timeout);
    $url =~ s/^(.*)\*$/$1/ if $Socket;
  }
  if (!$Socket && $proxy) {
    $address = $this->host_address($host);  $proxy = '';
  }
  $response{'Error-Message'} = 'Host not found' unless $address;
  return \%response unless $address;
  $response{'proxy'} = $proxy if $proxy; # Proxy used

  # Open the socket, connect to it and make it command buffered

  if (!$proxy) {
    $where = pack($sockaddr, AF_INET, $port, $address);
    $Socket = $this->Connect($where);
  }


  # Send the command to the server

  $url =~ tr/ /+/;  $object =~ tr/ /+/;
  my $StoredVersion = $this->{'versions'}->{$proxy ? $proxy : $host};
  if ($StoredVersion) {
    $version = $StoredVersion if $StoredVersion < $version;
  }
  my $command = $method . ' ' . ($proxy ? $url : $object) . ' HTTP/' . $version . $crlf;
  my $OK;
  my $i;
  my $previous = 1;
RETRY: for ($i = $previous;  $i <= $retry; $i++) {
  if (time() > $this->{'lifetime'}) {
    close($Socket);  $this->Clean();
    $response{'Error-Message'} = 'Lifetime expired';
    return \%response;
  }
  $OK = 1;
  my $old_pipe = $SIG{'PIPE'};
  $SIG{'PIPE'} = sub { die; };
  eval {
    my $h = $proxy ? $proxy : $host;
    $this->{'LastHost'} = $h;
    print $Socket $command;
    print $Socket 'Host: ', $h, $crlf;
    if (exists $request->{'headers'}) {
      $headers = $request->{'headers'};
      if (ref($headers) eq 'HASH') {
        foreach $k (keys %$headers) {
          print $Socket "$k: $$headers{$k}$crlf";
        }
      }
      elsif (ref($headers) eq 'ARRAY') {
        foreach $cmd (@$headers) {
          print $Socket $cmd, $crlf;
        }
      }
      else {
        print $Socket $headers, $crlf;
      }
    }
    if ($method eq 'POST') {
      print $Socket "Content-Type: $request->{'Content-Type'}$crlf";
      print $Socket "Content-Length: $request->{'Content-Length'}$crlf";
    }
    print $Socket $crlf;  # end of the command
    print $Socket $request->{'data'}, $crlf if $method eq 'POST';
  }; # End of EVAL
  $SIG{'PIPE'} = $old_pipe;
  $OK = 0 if $@;
  last if $OK;
  $Socket = $this->Reconnect() if $i < $retry;
} # End FOR
  $previous = $i + 1;
  if (!$OK) {
    my $alarm = $SIG{'ALRM'};
    $SIG{'ALRM'} = sub { die; };
    alarm{30};
    eval {
      close $Socket;
    };
    alarm(0);
    $SIG{'ALRM'} = $alarm;
    $this->Clean();
    $response{'Error-Message'} = 'Cannot connect';
    return \%response;
  }

  my $status = 0;

  do {

  # Read the headers and store them in the %response

  $response{'Status'} = 0;  # this will signal non-HTTP
  my $hhh = $crlf;
  while($hhh =~ /^\s*($crlf)*$/) {
    $hhh = $this->ReadBytes;
    last if !$this->{'socket'};
  };
  $_ = $hhh;
#                   1         2     3       4
  unless (/^HTTP\/(\S+)\s+(\d\d\d)(\S*)\s+(.*)$crlf/)
  {
    close $Socket;
    if ($previous <= $retry) {
      $Socket = $this->Reconnect();  goto RETRY;
    }
  }
  $response{'Status-Line'} = "HTTP/$1 $2$3 $4";
  $status = $response{'Status'} = $2;
  $response{'Sub-Status'} = $3;
  $response{'Protocol-Type'} = 'HTTP';
  my $rv = $response{'Protocol-Version'} = $1;
  $response{'Reason-Phrase'} = $4;
  if (!$StoredVersion || $StoredVersion < $rv) {
    $this->{'versions'}->{$proxy ? $proxy : $host} = $rv;
  }
  if ($status < 200) { # 100 Continue etc.
    while ($_ = $this->ReadBytes) {
      last if /^\w/;
    }
  }
  } until $status >= 200;
  my ($ReadHeaders, $fetched, $len) = (1, -1, 1);
  while ($_ = $this->ReadBytes) {
    if ($ReadHeaders && /^\w/) {
      if (/^(.*?):\s*(.*)\s*$crlf/) {
        my $field = $2;
        my $orig = lc $1;
        my $r;
        while ($orig =~ /^(.)(.*)$/) { # Canonical form of the response headers
          $r .= uc $1;
          $orig = $2;
          if ($orig =~ /^([^-]*)-(.*)$/) {
            $r .= $1 . '-';
            $orig = $2;
          }
          else {
            $r .= $orig;
            $orig = '';
          }
        }
        $r = 'ETag' if $r eq 'Etag';
        $r = 'Returned-Status' if $r eq 'Status';
        $response{$r} = $field;
      }
    }
    else {
      $ReadHeaders = 0;
      $len = $response{'Content-Length'};
      $fetched = 0;
      if ($status =~ /^2|406/ && $method =~ /^(GET|POST)$/) {
        if (!$len) {
          $len = 0x7fffffff;
        }
        if ($file =~ m!^(.+)/[^/]+$!) {
          mkpath $1, 0, 0777;
        }
        if ($status == 206) { # range
          my $range = $response{'Content-Range'};
          if ($range =~ /bytes\s+(\d+)-\d+\/\d+/) {
            $fetched = $1;
            if ($file) {
              $file = '+<' . $file unless $file =~ /^\+</;
              open(F, $file) or $file = 0;
              if ($file) {
                binmode(F);
                seek(F, 0, 2);
                $len = 0 unless $fetched <= tell(F);
                seek(F, $fetched, 0);
                my $fh = select(F);  $| = 1;  select($fh);
              }
              $fetched = 0;
            }
          }
          else {
            $len = 0;
          }
        }
        else {
          if ($file) {
            $file = '>' . $file unless $file =~ /^>/;
            open(F, $file) or $file = 0;
            if ($file) {
              binmode(F);
              my $fh = select(F);  $| = 1;  select($fh);
            }
          }
        }
        my ($rcvd, $x);
        if ($response{'Content-Type'} =~ /^text/) {
          my $chunk = '';
          while ($len && $len > 0) {
            if ($response{'Transfer-Encoding'} eq 'chunked') {
              $chunk = '';
              do {
                $x = $this->ReadBytes;
              } until $x =~ /^\s*[0-9a-fA-F]+\r?\n/;
              if ($x =~ /^\s*([0-9a-fA-F]+)\r?\n/) {
                $len = hex($1);
              }
              else { # invalid chunk -- this code will never be executed
                close($Socket);  $this->Clean();
                $response{'Error-Message'} = 'Invalid chunk';
                return \%response;
              }
            }
            $rcvd = 0;
            while (($len && $rcvd < $len) && ($x=$this->ReadBytes)) {
              $rcvd += length($x);
              if ($response{'Transfer-Encoding'} eq 'chunked') {
                my $y = $x;
                my $nn = chomp $x;
                $x =~ s/\r//g;
                $x = "$chunk$x";
                $chunk = substr($y, -$nn);
              }
              if ($file) {
                print F $x;
              }
              else { print $x; }
            }
            $fetched += $rcvd;
            $len = 0 unless $response{'Transfer-Encoding'} eq 'chunked';
          }
        }
        else {
          my $sz;
          while (($len && $fetched < $len) && ($x = $this->ReadBytes($len-$fetched))) {
            $fetched += length($x);
            if ($file) {
              print F $x;
            }
          }
        }
        close(F) if $file;
        last;
      }
      else {
        if ($len && $method =~ /^(GET|POST)$/) {
          my $s = 0;
          while ($s < $len) {
            my $x = $this->ReadBytes($len - $s);
            $s += length($x);
            last if length($x) <= 0;
          }
        }
        else {
          close $Socket;  $this->Clean();
        }
        last;
      }
    }
  }
  $response{'Bytes-Received'} = $fetched;
  $Socket = $this->{'socket'};  # may be already closed...
  if ($Socket &&
      ($response{'Protocol-Version'} < 1.1 || $version < 1.1 ||
      lc($response{'Connection'}) eq 'close' ||
      lc($request->{'Connection'}) eq 'close')) {
        close($Socket);  $this->Clean();
  }
  return \%response;
}

# Destruction

sub DESTROY {
my $this = shift;
$this->CloseConnection();
}
